"""The classes of messages supported by this socket impementation."""

from iotile.core.utilities.schema_verify import Verifier, NoneVerifier, DictionaryVerifier, StringVerifier
from iotile.core.utilities.schema_verify import LiteralVerifier, OptionsVerifier

# The prescribed schema of command response messages
# Messages with this format are automatically processed inside the Client
COMMAND = DictionaryVerifier()
COMMAND.add_required('type', LiteralVerifier('command'))
COMMAND.add_required('operation', StringVerifier())
COMMAND.add_required('uuid', StringVerifier())
COMMAND.add_optional('payload', Verifier())

SUCCESSFUL_RESPONSE = DictionaryVerifier()
SUCCESSFUL_RESPONSE.add_required('uuid', StringVerifier())
SUCCESSFUL_RESPONSE.add_required('type', LiteralVerifier('response'))
SUCCESSFUL_RESPONSE.add_required('success', LiteralVerifier(True))
SUCCESSFUL_RESPONSE.add_optional('payload', Verifier())

FAILURE_RESPONSE = DictionaryVerifier()
FAILURE_RESPONSE.add_required('type', LiteralVerifier('response'))
FAILURE_RESPONSE.add_required('uuid', StringVerifier())
FAILURE_RESPONSE.add_required('success', LiteralVerifier(False))
FAILURE_RESPONSE.add_required('reason', StringVerifier())
FAILURE_RESPONSE.add_required('exception_class', OptionsVerifier(StringVerifier(), NoneVerifier()))

RESPONSE = OptionsVerifier(SUCCESSFUL_RESPONSE, FAILURE_RESPONSE)

EVENT = DictionaryVerifier()
EVENT.add_required('type', LiteralVerifier('event'))
EVENT.add_required('name', StringVerifier())
EVENT.add_optional('payload', Verifier())

VALID_SERVER_MESSAGE = OptionsVerifier(RESPONSE, EVENT)
VALID_CLIENT_MESSAGE = OptionsVerifier(COMMAND)



from .errors import ServerCommandError
from .device_server import SocketDeviceServer
from .socket_server import AsyncSocketServer
from .device_adapter import SocketDeviceAdapter
from .socket_client import AsyncSocketClient
from .abstract_socket_implementation import AbstractSocketServerImplementation
from .abstract_socket_implementation import AbstractSocketClientImplementation

__all__ = ['ServerCommandError', 'SocketDeviceServer', 'AsyncSocketServer',
           'SocketDeviceAdapter', 'AsyncSocketClient', 'AbstractSocketServerImplementation',
           'AbstractSocketClientImplementation']



"""Generic implementation of serving access to a device over TCP sockets."""


from iotile.core.utilities import SharedLoop
from iotile_transport_socket_lib.generic import SocketDeviceServer
from .tcpsocket_implementation import TcpServerImplementation


_MISSING = object()

class TcpSocketDeviceServer(SocketDeviceServer):
    """A device server for connections to multiple devices via a SocketServer

    This class connects to an AbstractDeviceAdapter and serves it over
    TCP sockets. Currently the support arguments to pass in args are:

    - ``host``: The host name to serve on, defaults to 127.0.0.1
    - ``port``: The port name to serve on, defaults to a random port if not specified.
      If a random port is used, its value can be read on the ``port`` property after
      start() has completed.

    Args:
        adapter (AbstractDeviceAdapter): The device adapter that we should use
            to find devices.
        args (dict): Arguments to this device server.
        loop (BackgroundEventLoop): The background event loop we should
            run in.  Defaults to the shared global loop.
    """

    def __init__(self, adapter, args=None, *, loop=SharedLoop):
        host = args.get('host', '127.0.0.1')
        port_suggestion = args.get('port', 0)
        self.implementation = TcpServerImplementation(host, port_suggestion, loop)
        SocketDeviceServer.__init__(self, adapter, self.implementation, args, loop=loop)
        self.port = None



"""A server implementation matching AsyncValidatingWSClient."""

import logging
import asyncio

from iotile.core.exceptions import ValidationError

from iotile.core.utilities.async_tools import SharedLoop
from iotile_transport_socket_lib.protocol.messages import VALID_CLIENT_MESSAGE
from .packing import pack, unpack
from .errors import ServerCommandError


class _ConnectionContext:
    def __init__(self, server, con):
        self.server = server
        self.connection = con
        self.operations = set()
        self.user_data = None


class AsyncSocketServer:
    """A websocket server supporting command/response and server push.

    There is only one kind of message that can be sent from client to server:
    COMMAND.  Those messages are dispatched to command handler tasks that run
    in parallel as coroutines and send a RESPONSE message back when they are
    finished.

    The server can also, at any time, send an EVENT message to the client,
    which is able to register a callback for the event.

    This class is the server side implementation of AsyncSocketClient
    and is designed to be used with that class.

    Args:
        host (str): The host name on which we should register this server
        port (int): The port we should use.  If None is given (the default),
            then a random port is chosen and will be accessible on the
            port property after ``await start()`` finished without error.
        loop (BackgroundEventLoop): The background event loop we should
            run in.  Defaults to the shared global loop.
    """

    OPERATIONS_MAX_PRUNE = 10

    def __init__(self, implementation, loop=SharedLoop):
        self.task = None
        self.implementation = implementation

        self._commands = {}
        self._server_task = None
        self._loop = loop
        self._logger = logging.getLogger(__name__)

        logger = logging.getLogger('SocketServer')
        logger.setLevel(logging.ERROR)
        logger.addHandler(logging.NullHandler())

    async def prepare_conn(self, _con):
        """Called when a new connection is established.

        This method is designed to allow subclasses to store, per connection
        state.  The return value will be saved in a per connection context
        that is passed to all command coroutines.

        Args:
            _con (a connection object): The connection that we are supposed
                to prepare.

        Returns:
            object: Any user data that needs to be tracked with this connection.
        """

        return None

    async def teardown_conn(self, context):
        """Called when a connection is finalized.

        This is the partner call to prepare_conn() and is intended to give
        subclasses the ability to cleanly release any resources that were
        allocated along with this connection, outside of
        AsyncValidatingWSServer.

        Args:
            context (_ConnectionContext): An object with all of the
                associated data about the connection including any
                user data returned by prepare_conn().
        """

        pass

    def register_command(self, name, handler, validator):
        """Register a coroutine command handler.

        This handler will be called whenever a command message is received
        from the client, whose operation key matches ``name``.  The handler
        will be called as::

            response_payload = await handler(cmd_payload, context)

        If the coroutine returns, it will be assumed to have completed
        correctly and its return value will be sent as the result of the
        command.  If the coroutine wishes to signal an error handling the
        command, it must raise a ServerCommandError exception that contains a
        string reason code for the error.  This will generate an error
        response to the command.

        The cmd_payload is first verified using the SchemaVerifier passed in
        ``validator`` and handler is only called if verification succeeds. If
        verification fails, a failure response to the command is returned
        automatically to the client.

        Args:
            name (str): The unique command name that will be used to dispatch
                client command messages to this handler.
            handler (coroutine function): A coroutine function that will be
                called whenever this command is received.
            validator (SchemaVerifier): A validator object for checking the
                command payload before calling this handler.
        """

        self._commands[name] = (handler, validator)

    async def start(self):
        """Start the websocket server.

        When this method returns, the socket server will be running 

        This method should be called only once in the lifetime of the server
        and must be paired with a call to stop() to cleanly release the
        server's resources.
        """

        if self._server_task is not None:
            self._logger.debug("AsyncValidatingWSServer.start() called twice, ignoring")
            return

        started_signal = self._loop.create_future()
        self._server_task = self._loop.add_task(self._run_server_task(started_signal))
        await started_signal

    async def _run_server_task(self, started_signal):
        """Create a BackgroundTask to manage the server.

        This allows subclasess to attach their server related tasks as
        subtasks that are properly cleaned up when this parent task is
        stopped and not require them all to overload start() and stop()
        to perform this action.
        """

        server = await self.implementation.start_server(self._manage_connection, started_signal)
        if not server:
            return

        try:
            while True:
                await asyncio.sleep(1)
        except asyncio.CancelledError:
            self._logger.info("Stopping server due to stop() command")
        finally:
            server.close()
            await server.wait_closed()

        self._logger.debug("Server stopped, exiting task")

    async def stop(self):
        """Stop the websocket server.

        This method will shutdown the socket server and free all resources
        associated with it.  If there are ongoing connections at the time of
        shutdown, they will be cleanly closed.
        """

        if self._server_task is not None:
            await self._server_task.stop()
            self._server_task = None

    async def send_event(self, con, name, payload):
        """Send an event to a client connection.

        This method will push an event message to the client with the given
        name and payload.  You need to have access to the the ``connection``
        object for the client, which is only available once the client has
        connected and passed to self.prepare_conn(connection).

        Args:
            con (a connection object): The connection to use to send
                the event.
            name (str): The name of the event to send.
            payload (object): The msgpack-serializable object so send
                as the event's payload.
        """

        message = dict(type="event", name=name, payload=payload)
        encoded = pack(message)
        try:
            await self.implementation.send(con, encoded)
        except Exception:
            self._logger.debug("Error sending data")

    async def _manage_connection(self, con, _path):
        context = _ConnectionContext(self, con)

        try:
            try:
                context.user_data = await self.prepare_conn(con)
            except:
                self._logger.exception("Error preparing connection")
                raise

            while True:
                encoded = await self.implementation.recv(con)


                message = unpack(encoded)
                if not VALID_CLIENT_MESSAGE.matches(message):
                    self._logger.warning("Received invalid message %s", message)
                    continue

                handler = self._loop.get_loop().create_task(self._dispatch_message(con, message, context))
                context.operations.add(handler)

                if len(context.operations) > self.OPERATIONS_MAX_PRUNE:
                    _prune_finished(context.operations)
        except Exception:
            self._logger.debug("Error sending data")
        finally:
            await _cancel_operations(context.operations)

            try:
                await self.teardown_conn(context)
            except:  #pylint:disable=bare-except;This is a background worker
                self._logger.exception("Error tearing down connecion")
                raise

    async def _dispatch_message(self, con, message, context):
        cmd = message.get('operation')
        cmd_uuid = message.get('uuid')

        try:
            response = await self._try_call_command(message, context)
        except ServerCommandError as err:
            response = _error_response(err.reason, cmd_uuid)
        except NotImplementedError:
            self._logger.error("Command %s was not implemented", cmd)
            response = _error_response("Command %s was not implemented" % cmd, cmd_uuid)
        except Exception as err:  #pylint:disable=broad-except;We can't let a failing command take down the server
            self._logger.exception("Exception executing handler for command %s", cmd)

            if hasattr(err, 'reason'):
                reason = err.reason  #pylint:disable=no-member;We are explicitly checking for the member's existence
            elif hasattr(err, 'msg'):
                reason = err.msg  #pylint:disable=no-member;We are explicitly checking for the member's existence
            else:
                reason = str(err)

            response = _error_response(reason, cmd_uuid, exc=type(err).__name__)

        try:
            encoded_resp = pack(response)
        except:
            self._logger.exception("Unable to pack response message: %s", response)
            response = _error_response("Unable to pack response message", cmd_uuid)
            encoded_resp = pack(response)

        self._logger.debug("Sending response: %s", response)

        try:
            await self.implementation.send(con, encoded_resp)
        except ConnectionError:
            self._logger.debug("Response %s not sent because connection may be closed", response)

    async def _try_call_command(self, message, context):
        name = message.get('operation')
        uuid = message.get('uuid')
        payload = message.get('payload')

        self._logger.debug("Received command %s", message)

        handler_info = self._commands.get(name)
        if handler_info is None:
            raise ServerCommandError(name, 'Command %s not found' % name)

        handler, validator = handler_info

        try:
            payload = validator.verify(payload)
        except ValidationError as err:
            raise ServerCommandError(name, 'Invalid payload: %s' % err.params.get('reason'))

        payload = await handler(payload, context)
        return dict(type="response", uuid=uuid, success=True, payload=payload)


def _error_response(reason, uuid, exc=None):
    return dict(type="response", uuid=uuid, success=False,
                reason=reason, exception_class=exc)


def _prune_finished(operations):
    to_remove = [x for x in operations if x.done()]

    for operation in to_remove:
        operations.remove(operation)


async def _cancel_operations(operations):
    for handler in operations:
        if not handler.done():
            handler.cancel()

    await asyncio.gather(*list(operations), return_exceptions=True)



"""The Unix Socket Implementation that is used by the adapter and server"""

import logging
import asyncio
from iotile.core.utilities import SharedLoop
from iotile_transport_socket_lib.generic import AbstractSocketServerImplementation
from iotile_transport_socket_lib.generic import AbstractSocketClientImplementation
from iotile_transport_socket_lib.generic.abstract_socket_implementation import AsyncioSocketConnection

class UnixServerImplementation(AbstractSocketServerImplementation):
    """Unix flavor of a socket server

    Args:
        path (str): The path to the Unix Socket that will be opened. The file descriptor that it points
            to will be created by the asyncio.start_unix_server() call
        loop (iotile.core.utilities.BackgroundEventLoop): The background event loop we should
            run in.  Defaults to the shared global loop.
    """

    def __init__(self, path, loop=SharedLoop):
        self.path = path
        self._logger = logging.getLogger(__name__)
        self.loop = loop
        self._manage_connection_cb = None

    async def start_server(self, manage_connection_cb, started_signal):
        """Begin serving. Once started up, set the result of a given signal

        Args:
            manage_connection_cb (function): The function to call when a connection is made. Must accept
                a connection info object to store and later present back to this class to interact
                with that connection
            started_signal (asyncio.future): A future that will be set to True once the server has started
        """

        self._manage_connection_cb = manage_connection_cb
        try:
            server = await asyncio.start_unix_server(self._conn_cb_wrapper, path=self.path, loop=self.loop.get_loop())
            self._logger.debug("Serving on path %s", self.path)
            started_signal.set_result(True)
        except Exception as err:
            self._logger.exception("Error starting unix server on path %s", self.path)
            started_signal.set_exception(err)
            return
        return server

    async def _conn_cb_wrapper(self, reader, writer):
        """A wrapper function to encapsulate the result of a connection into a single object

            Unlike websockets which returns a single connection object, asyncio's servers return
            readers and writers. For simplicity, the reader and writers are stored in an object
            that provides read and write calls

        Args:
            reader (asyncio.StreamReader): The reader returned from the callback
            writer (asyncio.StreamWriter): The writer returned from the callback
        """

        unix_conn = AsyncioSocketConnection(reader, writer, self._logger)
        await self._manage_connection_cb(unix_conn, None)

    async def send(self, con, encoded):
        """Send the byte-encoded data to the given connection

        Args:
            con (AsyncioSocketConnection): The connection to send to
            encoded (bytes): Data to send
        """

        await con.send(encoded)

    async def recv(self, con):
        """Await incoming data from the given connection

        Args:
            con (AsyncioSocketConnection): The connection to receive from
        """
        return await con.recv()

class UnixClientImplementation(AbstractSocketClientImplementation):
    """Unix flavor of a Socket Connection

    Args:
        path (str): Path to the unix socket opened by the Server
        loop (iotile.core.utilities.BackgroundEventLoop): The background event loop we should
            run in.  Defaults to the shared global loop.
    """

    def __init__(self, path, loop=SharedLoop):
        self.path = path
        self._logger = logging.getLogger(__name__)
        self.loop = loop
        self.is_connected = False
        self.con = None

    async def connect(self):
        """Open the connection"""
        reader, writer = await asyncio.open_unix_connection(self.path, loop=self.loop.get_loop())
        self.con = AsyncioSocketConnection(reader, writer, self._logger)
        self._logger.debug("Connected to %s", self.path)

    async def close(self):
        """Close the connection"""
        await self.con.writer.close()
        self.con = None

    def connected(self):
        """Return true if there is an active connection"""
        return bool(self.con)

    async def send(self, encoded):
        """Send the bytes-encoded data

        Args:
            encoded (bytes): Data to send
        """
        await self.con.send(encoded)

    async def recv(self):
        """Await incoming data and return it"""
        return await self.con.recv()



"""Generic implementation of serving access to a device over websockets."""

import logging
import base64
from iotile.core.utilities import SharedLoop
from iotile.core.hw.transport.server import StandardDeviceServer
from iotile.core.hw.exceptions import VALID_RPC_EXCEPTIONS, DeviceServerError, DeviceAdapterError
from iotile.core.hw.virtual import pack_rpc_response
from iotile_transport_socket_lib.protocol import COMMANDS, OPERATIONS
from . import ServerCommandError
from .socket_server import AsyncSocketServer

_MISSING = object()

class SocketDeviceServer(StandardDeviceServer):
    """A device server for connections to multiple devices over any socket implementation

    Args:
        adapter (AbstractDeviceAdapter): The device adapter that we should use to find devices.
        implementation (AbstractSocketServer): The implementation of the socket Server
            that will be used to listen to connections
        loop (BackgroundEventLoop): The background event loop we should
            run in.  Defaults to the shared global loop.
    """

    def __init__(self, adapter, implementation, args=None, *, loop=SharedLoop):
        StandardDeviceServer.__init__(self, adapter, args, loop=loop)

        self.server = AsyncSocketServer(implementation, loop=loop)

        self.chunk_size = 4*1024  # Config chunk size to be 4kb for traces and reports streaming
        self._logger = logging.getLogger(__name__)

        self.server.register_command(OPERATIONS.CONNECT, self.connect_message, COMMANDS.ConnectCommand)
        self.server.register_command(OPERATIONS.PROBE, self.probe_message, COMMANDS.ProbeCommand)
        self.server.register_command(OPERATIONS.DISCONNECT, self.disconnect_message, COMMANDS.DisconnectCommand)
        self.server.register_command(OPERATIONS.OPEN_INTERFACE, self.open_interface_message,
                                     COMMANDS.OpenInterfaceCommand)
        self.server.register_command(OPERATIONS.CLOSE_INTERFACE, self.close_interface_message,
                                     COMMANDS.CloseInterfaceCommand)
        self.server.register_command(OPERATIONS.SEND_RPC, self.send_rpc_message, COMMANDS.SendRPCCommand)
        self.server.register_command(OPERATIONS.SEND_SCRIPT, self.send_script_message, COMMANDS.SendScriptCommand)
        self.server.register_command(OPERATIONS.DEBUG, self.debug_command_message, COMMANDS.SendDebugCommand)

        # Setup our hooks whenever a client connects or disconnects
        self.server.prepare_conn = self.prepare_conn
        self.server.teardown_conn = self.teardown_conn

    async def prepare_conn(self, conn):
        client_id = self.setup_client(user_data=conn, broadcast=True)

        self._logger.info("New client connection: %s", client_id)
        return client_id

    async def teardown_conn(self, context):
        client_id = context.user_data

        self._logger.info("Tearing down client connection: %s", client_id)
        await self.teardown_client(client_id)

    async def start(self):
        """Start serving access to devices.

        This method must not return until the devices are accessible over the
        implementations chosen server protocol.  It must be possible to
        immediately attach an AbstractDevice adapter to this server and use it
        without any race conditions.

        For example, if the server is configured to use a TCP port to serve
        devices, the TCP port must be bound and listening for connections when
        this coroutine returns.
        """

        await self.server.start()

    async def stop(self):
        """Stop serving access to devices.

        Subclass **must not** return until the devices are no longer accessible
        over the implementation's chosen server protocol.  There should be no
        race condition where a client is still able to invoke a function on
        this server after stop has returned.  This is important because server
        container running this AbstractDeviceServer may no longer be in a
        state to respond allow for client connections safely.

        Subclasses **must** cleanly disconnect from any devices accessed via
        ``adapter`` before stopping.  It is the job of the ``AbstractDeviceServer``
        to release all of its internal resources.  It is not the job of ``adapter``
        to somehow know that a particular DeviceServer has stopped and free the
        resources associated with that server.

        Subclasses **should** cleanly disconnect any clients when stopped if
        possible, rather than just breaking all sockets, for example.
        """

        await self.server.stop()

        # Cleanup any resources we had on the adapter
        await super(SocketDeviceServer, self).stop()

    async def probe_message(self, _message, context):
        """Handle a probe message.

        See :meth:`AbstractDeviceAdapter.probe`.
        """
        client_id = context.user_data

        await self.probe(client_id)

    async def connect_message(self, message, context):
        """Handle a connect message.

        See :meth:`AbstractDeviceAdapter.connect`.
        """
        conn_string = message.get('connection_string')
        client_id = context.user_data

        await self.connect(client_id, conn_string)

    async def disconnect_message(self, message, context):
        """Handle a disconnect message.

        See :meth:`AbstractDeviceAdapter.disconnect`.
        """

        conn_string = message.get('connection_string')
        client_id = context.user_data

        await self.disconnect(client_id, conn_string)

    async def open_interface_message(self, message, context):
        """Handle an open_interface message.

        See :meth:`AbstractDeviceAdapter.open_interface`.
        """

        conn_string = message.get('connection_string')
        interface = message.get('interface')
        client_id = context.user_data

        await self.open_interface(client_id, conn_string, interface)

    async def close_interface_message(self, message, context):
        """Handle a close_interface message.

        See :meth:`AbstractDeviceAdapter.close_interface`.
        """

        conn_string = message.get('connection_string')
        interface = message.get('interface')
        client_id = context.user_data

        await self.close_interface(client_id, conn_string, interface)

    async def send_rpc_message(self, message, context):
        """Handle a send_rpc message.

        See :meth:`AbstractDeviceAdapter.send_rpc`.
        """

        conn_string = message.get('connection_string')
        rpc_id = message.get('rpc_id')
        address = message.get('address')
        timeout = message.get('timeout')
        payload = message.get('payload')
        client_id = context.user_data

        self._logger.debug("Calling RPC %d:0x%04X with payload %s on %s",
                           address, rpc_id, payload, conn_string)

        response = bytes()
        err = None
        try:
            response = await self.send_rpc(client_id, conn_string, address, rpc_id, payload, timeout=timeout)
        except VALID_RPC_EXCEPTIONS as internal_err:
            err = internal_err
        except (DeviceAdapterError, DeviceServerError):
            raise
        except Exception as internal_err:
            self._logger.warning("Unexpected exception calling RPC %d:0x%04x", address, rpc_id, exc_info=True)
            raise ServerCommandError('send_rpc', str(internal_err)) from internal_err

        status, response = pack_rpc_response(response, err)
        return {
            'status': status,
            'payload': base64.b64encode(response)
        }

    async def send_script_message(self, message, context):
        """Handle a send_script message.

        See :meth:`AbstractDeviceAdapter.send_script`.
        """

        script = message.get('script')
        conn_string = message.get('connection_string')
        client_id = context.user_data

        if message.get('fragment_count') != 1:
            raise DeviceServerError(client_id, conn_string, 'send_script', 'fragmented scripts are not yet supported')

        await self.send_script(client_id, conn_string, script)

    async def debug_command_message(self, message, context):
        """Handle a debug message.

        See :meth:`AbstractDeviceAdapter.debug`.
        """

        conn_string = message.get('connection_string')
        command = message.get('command')
        args = message.get('args')
        client_id = context.user_data

        result = await self.debug(client_id, conn_string, command, args)
        return result

    async def client_event_handler(self, client_id, event_tuple, user_data):
        """Forward an event on behalf of a client.

        This method is called by StandardDeviceServer when it has an event that
        should be sent to a client.

        Args:
            client_id (str): The client that we should send this event to
            event_tuple (tuple): The conn_string, event_name and event
                object passed from the call to notify_event.
            user_data (object): The user data passed in the call to
                :meth:`setup_client`.
        """

        #TODO: Support sending disconnection events

        conn_string, event_name, event = event_tuple

        if event_name == 'report':
            report = event.serialize()
            report['encoded_report'] = base64.b64encode(report['encoded_report'])
            msg_payload = dict(connection_string=conn_string, serialized_report=report)
            msg_name = OPERATIONS.NOTIFY_REPORT
        elif event_name == 'trace':
            encoded_payload = base64.b64encode(event)
            msg_payload = dict(connection_string=conn_string, payload=encoded_payload)
            msg_name = OPERATIONS.NOTIFY_TRACE
        elif event_name == 'progress':
            msg_payload = dict(connection_string=conn_string, operation=event.get('operation'),
                               done_count=event.get('finished'), total_count=event.get('total'))
            msg_name = OPERATIONS.NOTIFY_PROGRESS
        elif event_name == 'device_seen':
            msg_payload = event
            msg_name = OPERATIONS.NOTIFY_DEVICE_FOUND
        elif event_name == 'broadcast':
            report = event.serialize()
            report['encoded_report'] = base64.b64encode(report['encoded_report'])
            msg_payload = dict(connection_string=conn_string, serialized_report=report)
            msg_name = OPERATIONS.NOTIFY_BROADCAST
        else:
            self._logger.debug("Not forwarding unknown event over websockets: %s", event_tuple)
            return

        try:
            self._logger.debug("Sending event %s: %s", msg_name, msg_payload)
            await self.server.send_event(user_data, msg_name, msg_payload)
        except ConnectionError:
            self._logger.debug("Could not send notification because connection was closed for client %s", client_id)


"""List of defined commands and events."""

# Commands
CONNECT = 'connect'
CLOSE_INTERFACE = 'close_interface'
OPEN_INTERFACE = 'open_interface'
PROBE = 'probe'
SEND_RPC = 'send_rpc'
SEND_SCRIPT = 'send_script'
DEBUG = 'debug_command'
DISCONNECT = 'disconnect'

COMMANDS = frozenset([CONNECT, CLOSE_INTERFACE, OPEN_INTERFACE, PROBE, SEND_RPC,
                      SEND_SCRIPT, DISCONNECT, DEBUG])

# Events
NOTIFY_DEVICE_FOUND = 'device_found'
NOTIFY_PROGRESS = 'progress'
NOTIFY_REPORT = 'report'
NOTIFY_BROADCAST = 'broadcast'
NOTIFY_TRACE = 'trace'
NOTIFY_DISCONNECT = 'unexpected_disconnect'
NOTIFY_SOCKET_DISCONNECT = 'unexpected_socket_disconnect'

EVENTS = frozenset([NOTIFY_DEVICE_FOUND, NOTIFY_PROGRESS, NOTIFY_REPORT,
                    NOTIFY_BROADCAST, NOTIFY_TRACE, NOTIFY_DISCONNECT, NOTIFY_SOCKET_DISCONNECT])



"""Presents a SocketDeviceServer that serves via Async Unix Sockets."""

from iotile.core.utilities import SharedLoop
from iotile_transport_socket_lib.generic import SocketDeviceServer
from .unixsocket_implementation import UnixServerImplementation

class UnixSocketDeviceServer(SocketDeviceServer):
    """A device server for connections to multiple devices via a SocketServer

    This class connects to an AbstractDeviceAdapter and serves it over
    Unix Sockets. Currently the support arguments to pass in args are:

    - ``path``: The path to the Unix Socket that will be opened. The file descriptor that it points
        to will be created by the UnixServerImplementation's asyncio.start_unix_server() call

    Args:
        adapter (AbstractDeviceAdapter): The device adapter that we should use
            to find devices.
        args (dict): Arguments to this device server.
        loop(iotile.core.utilities.BackgroundEventLoop): The background event loop we should
            run in.  Defaults to the shared global loop.
    """

    def __init__(self, adapter, args=None, *, loop=SharedLoop):
        path = args.get('path', None)
        self.implementation = UnixServerImplementation(path, loop)
        SocketDeviceServer.__init__(self, adapter, self.implementation, args, loop=loop)



# This file is copyright Arch Systems, Inc.
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

import base64
import logging
import asyncio
from iotile.core.hw.transport.adapter import StandardDeviceAdapter
from iotile.core.utilities.schema_verify import NoneVerifier
from iotile.core.utilities import SharedLoop
from iotile.core.hw.virtual import unpack_rpc_response
from iotile.core.hw.reports import IOTileReportParser
from iotile.core.hw.exceptions import DeviceAdapterError
from iotile.core.exceptions import ExternalError
from iotile_transport_socket_lib.protocol import OPERATIONS, NOTIFICATIONS, COMMANDS
from .socket_client import AsyncSocketClient

class SocketDeviceAdapter(StandardDeviceAdapter):
    """ A device adapter allowing connections to devices over any socket implementation

    Args:
        implementation (AbstractSocketClient): The implementation of the socket client
            that will be used to connect to the SocketDeviceServer
        loop (BackgroundEventLoop): Loop for running our websocket client.
    """

    def __init__(self, implementation, *, loop=SharedLoop):
        super(SocketDeviceAdapter, self).__init__(loop=loop)

        # Configuration
        self.set_config('default_timeout', 10.0)
        self.set_config('expiration_time', 60.0)
        self.set_config('max_connections', 100)
        self.set_config('probe_required', True)
        self.set_config('probe_supported', True)

        # Set logger
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.NullHandler())

        self._report_parser = IOTileReportParser()

        self.client = AsyncSocketClient(implementation, loop=loop)
        self.client.register_event(OPERATIONS.NOTIFY_DEVICE_FOUND, self._on_device_found,
                                   NOTIFICATIONS.ScanEvent)
        self.client.register_event(OPERATIONS.NOTIFY_TRACE, self._on_trace_notification,
                                   NOTIFICATIONS.TraceEvent)
        self.client.register_event(OPERATIONS.NOTIFY_REPORT, self._on_report_notification,
                                   NOTIFICATIONS.ReportEvent)
        self.client.register_event(OPERATIONS.NOTIFY_BROADCAST, self._on_broadcast_notification,
                                   NOTIFICATIONS.ReportEvent)
        self.client.register_event(OPERATIONS.NOTIFY_PROGRESS, self._on_progress_notification,
                                   NOTIFICATIONS.ProgressEvent)
        self.client.register_event(OPERATIONS.NOTIFY_SOCKET_DISCONNECT,
                                   self._on_socket_disconnect, NoneVerifier())

    async def start(self):
        """Start the device adapter.

        See :meth:`AbstractDeviceAdapter.start`.
        """

        await self.client.start()

    async def stop(self):
        """Stop the device adapter.

        See :meth:`AbstractDeviceAdapter.stop`.
        """

        await self.client.stop()

    async def probe(self):
        """Probe for devices connected to this adapter.

        See :meth:`AbstractDeviceAdapter.probe`.
        """

        await self._send_command(OPERATIONS.PROBE, None, COMMANDS.ProbeResponse)

    async def connect(self, conn_id, connection_string):
        """Connect to a device.

        See :meth:`AbstractDeviceAdapter.connect`.
        """

        self._ensure_connection(conn_id, False)

        msg = dict(connection_string=connection_string)
        await self._send_command(OPERATIONS.CONNECT, msg, COMMANDS.ConnectResponse)

        self._setup_connection(conn_id, connection_string)

    async def disconnect(self, conn_id):
        """Disconnect from a connected device.

        See :meth:`AbstractDeviceAdapter.disconnect`.
        """

        self._ensure_connection(conn_id, True)

        msg = dict(connection_string=self._get_property(conn_id, "connection_string"))

        try:
            await self._send_command(OPERATIONS.DISCONNECT, msg, COMMANDS.DisconnectResponse)
        finally:
            self._teardown_connection(conn_id)

    async def open_interface(self, conn_id, interface):
        """Open an interface on an IOTile device.

        See :meth:`AbstractDeviceAdapter.open_interface`.
        """

        self._ensure_connection(conn_id, True)
        connection_string = self._get_property(conn_id, "connection_string")

        msg = dict(interface=interface, connection_string=connection_string)
        await self._send_command(OPERATIONS.OPEN_INTERFACE, msg, COMMANDS.OpenInterfaceResponse)

    async def close_interface(self, conn_id, interface):
        """Close an interface on this IOTile device.

        See :meth:`AbstractDeviceAdapter.close_interface`.
        """

        self._ensure_connection(conn_id, True)
        connection_string = self._get_property(conn_id, "connection_string")

        msg = dict(interface=interface, connection_string=connection_string)
        await self._send_command(OPERATIONS.CLOSE_INTERFACE, msg, COMMANDS.CloseInterfaceResponse)

    async def send_rpc(self, conn_id, address, rpc_id, payload, timeout):
        """Send an RPC to a device.

        See :meth:`AbstractDeviceAdapter.send_rpc`.
        """

        self._ensure_connection(conn_id, True)
        connection_string = self._get_property(conn_id, "connection_string")

        msg = dict(address=address, rpc_id=rpc_id, payload=base64.b64encode(payload),
                   timeout=timeout, connection_string=connection_string)

        response = await self._send_command(OPERATIONS.SEND_RPC, msg, COMMANDS.SendRPCResponse,
                                            timeout=timeout)

        return unpack_rpc_response(response.get('status'), response.get('payload'),
                                   rpc_id=rpc_id, address=address)

    async def send_script(self, conn_id, data):
        """Send a a script to this IOTile device

        Args:
            conn_id (int): A unique identifier that will refer to this connection
            data (bytes): the script to send to the device
        """

        self._ensure_connection(conn_id, True)
        connection_string = self._get_property(conn_id, "connection_string")

        msg = dict(connection_string=connection_string, fragment_count=1, fragment_index=0,
                   script=base64.b64encode(data))
        await self._send_command(OPERATIONS.SEND_SCRIPT, msg, COMMANDS.SendScriptResponse)

    async def _on_device_found(self, device):
        """Callback function called when a new device has been scanned by the probe.

        Args:
            response (dict): The device advertisement data
        """

        await self.notify_event(device.get('connection_string'), 'device_seen', device)

    async def _on_report_notification(self, event):
        """Callback function called when a report event is received.

        Args:
            event (dict): The report_event
        """

        conn_string = event.get('connection_string')
        report = self._report_parser.deserialize_report(event.get('serialized_report'))

        self.notify_event(conn_string, 'report', report)

    async def _on_broadcast_notification(self, event):
        """Callback function called when a broadcast event is received.

        Args:
            event (dict): The broadcast event
        """

        conn_string = event.get('connection_string')
        report = self._report_parser.deserialize_report(event.get('serialized_report'))

        self.notify_event(conn_string, 'broadcast', report)

    async def _on_trace_notification(self, trace_event):
        """Callback function called when a trace chunk is received.

        Args:
            trace_chunk (dict): The received trace chunk information
        """

        conn_string = trace_event.get('connection_string')
        payload = trace_event.get('payload')

        await self.notify_event(conn_string, 'trace', payload)

    async def _on_progress_notification(self, progress):
        """Callback function called when a progress notification is received.

        Args:
            progress (dict): The received notification containing the progress information
        """

        conn_string = progress.get('connection_string')
        done = progress.get('done_count')
        total = progress.get('total_count')
        operation = progress.get('operation')

        await self.notify_progress(conn_string, operation, done, total, wait=True)

    async def _on_socket_disconnect(self, _event):
        """Callback function called when we have been disconnected from the server (by error or not).
        Allows to clean all if the disconnection was unexpected."""

        self.logger.info('Forcibly disconnected from the socket server')

        conns = self._connections.copy()
        for conn_id in conns:
            conn_string = self._get_property(conn_id, 'connection_string')
            self._teardown_connection(conn_id)
            self.notify_event(conn_string, 'disconnect', "Socket connection closed")

    async def _send_command(self, name, args, verifier, timeout=10.0):
        try:
            return await self.client.send_command(name, args, verifier, timeout=timeout)
        except ExternalError as err:
            raise DeviceAdapterError(None, name, err.params['reason'])
        except asyncio.TimeoutError as err:
            raise DeviceAdapterError(None, name, 'operation timed out after %f seconds' % timeout)



"""Presents a SocketDeviceAdapter that connects via Async Unix Sockets."""

from iotile.core.utilities import SharedLoop

from iotile_transport_socket_lib.generic import SocketDeviceAdapter
from .unixsocket_implementation import UnixClientImplementation

class UnixSocketDeviceAdapter(SocketDeviceAdapter):
    """A device adapter allowing connections to devices over Unix Sockets.

    Args:
        port (str): The path to the unix socket opened by the Device Server
            ex: "/tmp/gateway_socket". Note: This is usually passed in via config file
            in the format "unix:<port>". The "unix:" part is parsed and dropped by the
            HardwareManager
        loop(iotile.core.utilities.BackgroundEventLoop): The background event loop we should
            run in.  Defaults to the shared global loop.
    """

    def __init__(self, port, *, loop=SharedLoop):
        self.implementation = UnixClientImplementation(port, loop)
        SocketDeviceAdapter.__init__(self, self.implementation, loop=loop)



"""List of notifications handled by the Socket plugin."""

from iotile.core.utilities.schema_verify import BytesVerifier, DictionaryVerifier, Verifier, BooleanVerifier, IntVerifier,\
    StringVerifier


# Device found while scanning
ScanEvent = Verifier()

# Report
SerializedReport = DictionaryVerifier()
SerializedReport.add_required('encoded_report', BytesVerifier(encoding="base64"))
SerializedReport.add_required('received_time', Verifier())
SerializedReport.add_required('report_format', IntVerifier())
SerializedReport.add_required('origin', IntVerifier())

ReportEvent = DictionaryVerifier()
ReportEvent.add_required('connection_string', StringVerifier())
ReportEvent.add_required('serialized_report', SerializedReport)


DisconnectionEvent = DictionaryVerifier()
DisconnectionEvent.add_required('connection_string', StringVerifier())
DisconnectionEvent.add_required('reason', StringVerifier())
DisconnectionEvent.add_required('expected', BooleanVerifier())

# Trace
TraceEvent = DictionaryVerifier()
TraceEvent.add_required('connection_string', StringVerifier())
TraceEvent.add_required('payload', BytesVerifier(encoding="base64"))

# Script and debug progress
ProgressEvent = DictionaryVerifier()
ProgressEvent.add_required('connection_string', StringVerifier())
ProgressEvent.add_required('operation', StringVerifier())
ProgressEvent.add_required('done_count', IntVerifier())
ProgressEvent.add_required('total_count', IntVerifier())



"""List of known command and response payloads."""

from iotile.core.utilities.schema_verify import BytesVerifier, DictionaryVerifier, \
    EnumVerifier, FloatVerifier, IntVerifier, StringVerifier, NoneVerifier, Verifier, \
    OptionsVerifier

# Connect Command
ConnectCommand = DictionaryVerifier()
ConnectCommand.add_required('connection_string', StringVerifier())

ConnectResponse = NoneVerifier()

# Disconnect Command
DisconnectCommand = DictionaryVerifier()
DisconnectCommand.add_required('connection_string', StringVerifier())

DisconnectResponse = NoneVerifier()


_InterfaceEnum = EnumVerifier(['rpc', 'streaming', 'tracing', 'script', 'debug'])

# OpenInterface Command

OpenInterfaceCommand = DictionaryVerifier()
OpenInterfaceCommand.add_required('interface', _InterfaceEnum)
OpenInterfaceCommand.add_required('connection_string', StringVerifier())

OpenInterfaceResponse = NoneVerifier()
# CloseInterface Command

CloseInterfaceCommand = DictionaryVerifier()
CloseInterfaceCommand.add_required('interface', _InterfaceEnum)
CloseInterfaceCommand.add_required('connection_string', StringVerifier())

CloseInterfaceResponse = NoneVerifier()

# Probe
ProbeCommand = NoneVerifier()
ProbeResponse = NoneVerifier()

# Send RPC
SendRPCCommand = DictionaryVerifier()
SendRPCCommand.add_required('connection_string', StringVerifier())
SendRPCCommand.add_required('address', IntVerifier())
SendRPCCommand.add_required('rpc_id', IntVerifier())
SendRPCCommand.add_required('timeout', FloatVerifier())
SendRPCCommand.add_required('payload', BytesVerifier(encoding="base64"))

SendRPCResponse = DictionaryVerifier()
SendRPCResponse.add_required('status', IntVerifier())
SendRPCResponse.add_required('payload', BytesVerifier(encoding="base64"))

# Send script
SendScriptCommand = DictionaryVerifier()
SendScriptCommand.add_required('connection_string', StringVerifier())
SendScriptCommand.add_required('fragment_count', IntVerifier())
SendScriptCommand.add_required('fragment_index', IntVerifier())
SendScriptCommand.add_required('script', BytesVerifier(encoding="base64"))

SendScriptResponse = NoneVerifier()

SendDebugCommand = DictionaryVerifier()
SendScriptCommand.add_required('connection_string', StringVerifier())
SendDebugCommand.add_required('command', StringVerifier())
SendDebugCommand.add_required('args', Verifier())

SendDebugResponse = Verifier()



# This file is copyright Arch Systems, Inc.
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

from iotile.core.utilities import SharedLoop

from iotile_transport_socket_lib.generic import SocketDeviceAdapter
from .tcpsocket_implementation import TcpClientImplementation


class TcpSocketDeviceAdapter(SocketDeviceAdapter):
    """A device adapter allowing connections to devices over WebSockets.

    Args:
        port (string): A target for the TCP Socket client to connect to in form of
            server:port.  For example, "localhost:5120".
        loop (iotile.core.utilities.BackgroundEventLoop): The background event loop we should
            run in.  Defaults to the shared global loop.
    """

    def __init__(self, port, *, loop=SharedLoop):
        host = port.split(':')[0]
        tcp_port = port.split(':')[1]
        self.implementation = TcpClientImplementation(host, tcp_port, loop)

        SocketDeviceAdapter.__init__(self, self.implementation, loop=loop)



"""Abstract base class for different socket transport implementations

Allows the AbstractDeviceAdapter and Servers to communicate via any different
protocols depending onsystem architect's choice

The functions will contain implementation details specific to the protocol.
For example, Unix Socket's send and receive functions will encode and send the
size of each transmission first and call the asyncio StreamReader and Writer
send and receive functions.
"""

import abc
import struct
import asyncio


class AsyncioSocketConnection:
    """Paired read and write commands for an active Socket Connection

    Args:
        reader (asyncio.StreamReader): The reader returned from the connect command or callback
        writer (asyncio.StreamWriter): The writer returned from the connect command or callback
        logger (logging.Logger): The logger to use (Since this is used by both the Server and Client, the
            logger should be that of the Server or Client)
    """

    _ARCHMAGIC = "ARCH".encode()
    _HEADERFORMAT = '4sL'
    def __init__(self, reader, writer, logger):
        self.reader = reader
        self.writer = writer
        self._logger = logger

    async def send(self, encoded):
        """Send encoded byte data. First send a header with the data's length, then the data itself
        Return once data in the buffer has been drained

        Args:
            encoded (bytes): Encoded data to send
        """
        try:
            packed_header = struct.pack(self._HEADERFORMAT, self._ARCHMAGIC, len(encoded))
            await self.writer.drain()
            self.writer.write(packed_header)
            self.writer.write(encoded)

        except Exception:
            raise ConnectionError from Exception

    async def recv(self):
        """Await incoming data, return the encoded bytes

        Returns:
            bytes: Encoded data object
        """
        packed_header = await self.reader.read(struct.calcsize(self._HEADERFORMAT))
        magic, encoded_len = struct.unpack(self._HEADERFORMAT, packed_header)

        if magic != self._ARCHMAGIC:
            raise struct.error

        encoded = await self.reader.read(encoded_len)
        return encoded


class AbstractSocketServerImplementation(abc.ABC):
    """Abstract Socket Server"""

    @abc.abstractmethod
    def start_server(self, manage_connection_cb, started_signal):
        """Begin serving. Once started up, set the result of a given signal

        Args:
            manage_connection_cb (function): The function to call when a connection is made. Must accept
                a connection info object to store and later present back to this class to interact
                with that connection
            started_signal (asyncio.future): A future that will be set once the server is started.
        """

    @abc.abstractmethod
    async def send(self, con, encoded):
        """Send the byte-encoded data to the given connection

        Args:
            con (a connection object): The connection to send to. The implementation
                must know how to use this object to send the data to the right target
            encoded (bytes): Data to send
        """

    @abc.abstractmethod
    async def recv(self, con):
        """Await incoming data from the given connection

        Args:
            con (a connection object): The connection to receive from.
                The implementation must know how to use this object
        """

class AbstractSocketClientImplementation(abc.ABC):
    """Abstract Socket Client"""

    @abc.abstractmethod
    async def connect(self):
        """Open the connection"""

    @abc.abstractmethod
    async def close(self):
        """Close the connection"""

    @abc.abstractmethod
    def connected(self):
        """Return true if there is an active connection"""

    @abc.abstractmethod
    async def send(self, packed):
        """Send the bytes-encoded data

        Args:
            packed(bytes): Data to send
        """

    @abc.abstractmethod
    async def recv(self):
        """Await incoming data and return it"""
        


from setuptools import setup, find_packages

import version

setup(
    name="iotile-transport-socket-lib",
    packages=find_packages(exclude=("test",)),
    version=version.version,
    license="LGPLv3",
    install_requires=[
        "iotile-core>=5.2",
        "msgpack>=1",
    ],
    python_requires=">=3.7,<4",
    entry_points={
        'iotile.device_adapter': [
            'unix = iotile_transport_socket_lib.unix_socket:UnixSocketDeviceAdapter',
            'tcp = iotile_transport_socket_lib.tcp_socket:TcpSocketDeviceAdapter'
        ],
        'iotile.device_server': [
            'unixsocket = iotile_transport_socket_lib.unix_socket:UnixSocketDeviceServer',
            'tcpsocket = iotile_transport_socket_lib.tcp_socket:TcpSocketDeviceServer'
        ]
    },
    description="IOTile Transport Socket Library Plugin",
    author="Arch",
    author_email="info@arch-iot.com",
    url="http://github.com/iotile/coretools",
    keywords=["iotile", "arch", "embedded", "hardware", "firmware"],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
        ],
    long_description="""\
IOTile Transport Socket Library Plugin
--------------------------------------

A python plug for IOTile CoreTools that provides common message formats, packing routines and
other code that is useful for performing iotile operations via socket or 
socket-like protocols. Not particularily useful on it's own.
"""
)


"""Exceptions specific to the sockets subpackage."""

class ServerCommandError(Exception):
    """An internal exception used to signal failure be a command handler.

    This exception communicates from a command handler to the server that the command failed and triggers an
    error response.

    Args:
        command (str): The name of the command that failed.
        reason (str): The reason the command failed.
    """

    def __init__(self, command, reason):
        super(ServerCommandError, self).__init__()

        self.command = command
        self.reason = reason



import logging
import uuid
import inspect
import asyncio
from iotile.core.exceptions import ExternalError, ValidationError
from iotile.core.utilities.async_tools import OperationManager, SharedLoop
from iotile.core.utilities.schema_verify import Verifier
from iotile_transport_socket_lib.generic.packing import pack, unpack
from iotile_transport_socket_lib.protocol.messages import VALID_SERVER_MESSAGE
from iotile_transport_socket_lib.protocol.operations import NOTIFY_SOCKET_DISCONNECT

class AsyncSocketClient:
    """An asynchronous socket client that validates messages received.

    This client is designed to allow the easy construction of client/server
    protocols where the client can send commands to the server that
    receive responses and the server can push events to the client at
    any time.

    The schema of all messages exchanged are validated according to
    SchemaVerifiers to ensure they are correct before passing on to
    the underlying handlers.

    Messages are packed using msgpack in a binary format and are decoded and
    validated automatically.

    Args:
        implementation (AbstractSocketClient): The implementation of the socket client
            that will be used for data transport
        loop (BackgroundEventLoop): The background event loop that we should
            run in, or None to use the default shared background loop.
        logger_name (str): Optional name for the logger we should use to
            log messages.
    """

    def __init__(self, implementation, loop=SharedLoop, logger_name=__name__):
        self._implementation = implementation

        self._connection_task = None
        self._logger = logging.getLogger(logger_name)
        self._loop = loop
        self._event_validators = {}
        self._allowed_exceptions = {}
        self._manager = OperationManager(loop=loop)

        logger = logging.getLogger('SocketClient')
        logger.setLevel(logging.ERROR)
        logger.addHandler(logging.NullHandler())

    def allow_exception(self, exc_class):
        """Allow raising this class of exceptions from commands.

        When a command fails on the server side due to an exception, by
        default it is turned into a string and raised on the client side as an
        ExternalError.  The original class name is sent but ignored.  If you
        would like to instead raise an instance of the same exception on the
        client side, you can pass the exception class object to this method
        and instances of that exception will be reraised.

        The caveat is that the exception must be creatable with a single
        string parameter and it should have a ``msg`` property.

        Args:
            exc_class (class): A class object with the exception that
                we should allow to pass from server to client.
        """

        name = exc_class.__name__
        self._allowed_exceptions[name] = exc_class

    async def start(self, name="socket_client"):
        """Connect to the socket server.

        This method will spawn a background task in the designated event loop
        that will run until stop() is called.  You can control the name of the
        background task for debugging purposes using the name parameter.  The
        name is not used in anyway except for debug logging statements.

        Args:
            name (str): Optional name for the background task.
        """

        await self._implementation.connect()
        self._connection_task = self._loop.add_task(self._manage_connection(), name=name)

    async def stop(self):
        """Stop this socket client and disconnect from the server.

        This method is idempotent and may be called multiple times.  If called
        when there is no active connection, it will simply return.
        """

        if self._connection_task is None:
            return

        try:
            await self._connection_task.stop()
        finally:
            self._connection_task = None
            self._manager.clear()

    async def send_command(self, command, args, validator, timeout=10.0):
        """Send a command and synchronously wait for a single response.

        Args:
            command (string): The command name
            args (dict): Optional arguments.
            validator (Verifier): A SchemaVerifier to verify the response
                payload.
            timeout (float): The maximum time to wait for a response.
                Defaults to 10 seconds.

        Returns:
            dict: The response payload

        Raises:
            ExternalError: If the server is not connected or the command
                fails.
            asyncio.TimeoutError: If the command times out.
            ValidationError: If the response payload does not match the
                given validator.
        """

        if not self._implementation.connected:
            raise ExternalError("No websock connection established")

        cmd_uuid = str(uuid.uuid4())
        msg = dict(type='command', operation=command, uuid=cmd_uuid,
                   payload=args)

        packed = pack(msg)

        # Note: register future before sending to avoid race conditions
        response_future = self._manager.wait_for(type="response", uuid=cmd_uuid,
                                                 timeout=timeout)

        await self._implementation.send(packed)

        response = await response_future

        if response.get('success') is False:
            self._raise_error(command, response)

        if validator is None:
            return response.get('payload')

        return validator.verify(response.get('payload'))

    def _raise_error(self, command, response):
        exc_name = response.get('exception_class')
        reason = response.get('reason')

        exc_class = self._allowed_exceptions.get(exc_name)
        if exc_class is not None:
            raise exc_class(reason)

        raise ExternalError("Command {} failed".format(command),
                            reason=response.get('reason'))

    async def _manage_connection(self):
        """Internal coroutine for managing the client connection."""

        try:
            while True:
                message = await self._implementation.recv()

                try:
                    unpacked = unpack(message)
                except Exception:  # pylint:disable=broad-except;This is a background worker
                    self._logger.exception("Corrupt message received")
                    continue

                if not VALID_SERVER_MESSAGE.matches(unpacked):
                    self._logger.warning("Dropping invalid message from server: %s", unpacked)
                    continue

                # Don't block until all callbacks have finished since once of
                # those callbacks may call self.send_command, which would deadlock
                # since it couldn't get the response until it had already finished.
                if not await self._manager.process_message(unpacked, wait=False):
                    self._logger.warning("No handler found for received message, message=%s", unpacked)
        except asyncio.CancelledError:
            self._logger.info("Closing connection to server due to stop()")
        finally:
            await self._manager.process_message(dict(type='event', name=NOTIFY_SOCKET_DISCONNECT, payload=None))
            await self._implementation.close()

    def register_event(self, name, callback, validator):
        """Register a callback to receive events.

        Every event with the matching name will have its payload validated
        using validator and then will be passed to callback if validation
        succeeds.

        Callback must be a normal callback function, coroutines are not
        allowed.  If you need to run a coroutine you are free to schedule it
        from your callback.

        Args:
            name (str): The name of the event that we are listening
                for
            callback (callable): The function that should be called
                when a message that matches validator is received.
            validator (Verifier): A schema verifier that will
                validate a received message uniquely
        """

        async def _validate_and_call(message):
            payload = message.get('payload')

            try:
                payload = validator.verify(payload)
            except ValidationError:
                self._logger.warning("Dropping invalid payload for event %s, payload=%s",
                                     name, payload)
                return

            try:
                result = callback(payload)
                if inspect.isawaitable(result):
                    await result
            except:  # pylint:disable=bare-except;This is a background logging routine
                self._logger.error("Error calling callback for event %s, payload=%s",
                                   name, payload, exc_info=True)

        self._manager.every_match(_validate_and_call, type="event", name=name)

    def post_command(self, command, args):
        """Post a command asynchronously and don't wait for a response.

        There is no notification of any error that could happen during
        command execution.  A log message will be generated if an error
        occurred.  The command's response is discarded.

        This method is thread-safe and may be called from inside or ouside
        of the background event loop.  If there is no connection,
        no error will be raised (though an error will be logged).

        Args:
            command (string): The command name
            args (dict): Optional arguments
        """

        self._loop.log_coroutine(self.send_command(command, args, Verifier()))



"""Helper functions for packing/unpacking msgpack messages."""

import datetime
import msgpack


def unpack(message):
    """Unpack a binary msgpacked message."""

    return msgpack.unpackb(message, object_hook=_decode_datetime)


def pack(message):
    """Pack a message into a binary packed message with datetime handling."""

    return msgpack.packb(message, default=_encode_datetime)


def _decode_datetime(obj):
    """Decode a msgpack'ed datetime."""

    if '__datetime__' in obj:
        obj = datetime.datetime.strptime(obj['as_str'].decode(), "%Y%m%dT%H:%M:%S.%f")
    return obj


def _encode_datetime(obj):
    """Encode a msgpck'ed datetime."""

    if isinstance(obj, datetime.datetime):
        obj = {'__datetime__': True, 'as_str': obj.strftime("%Y%m%dT%H:%M:%S.%f").encode()}
    return obj
